# T_SKYCAL -- Read reference features, shift, and stretch
# June 25, 2003
# By Doug Mink

# SKYCAL reads a file produced by XCSAO containing the results of correlations
# between a template and a spectrum which is the same as the template but
# with distortions.  It then fits the distortions and creates a database/id
# file using the centers of the cross-correlated regions as the lines.

include <pkg/gtools.h>
include "identify.h"

define  MAXPTS          256

procedure t_skycal

char	refin[SZ_FNAME]		# Input reference image
char	refout[SZ_FNAME]	# Output reference image
char	shiftfile[SZ_FNAME]	# XCSAO output file containing pixels and shifts
char	dbfile[SZ_FNAME]	# Path to id database file

bool	verbose, debug
int	i, nfeatures
int	ap[2]
double	dindef, dline0
char	s1[32], s2[32], s3[32]
real	d1, d2, d3, d4
double	x[MAXPTS], y[MAXPTS], wts[MAXPTS]
int	ir

char	fitfunc[SZ_FNAME]	# Continuum fitting function type
int	fitorder		# order of fit
int	niterate		# number of iterations for fit
real	grow			# radius around rejected points to replace
bool	interact		# Use interactive fitting (true or false)
int	npts
real	hirej, lowrej
int	fd
pointer	cv
pointer	ic
pointer	gopen(), gt_init()
int     fscan(), open()

int	clgeti()
bool	clgetb()
#double	clgetd()
double	dcveval()
pointer	id		# ID pointer
pointer	gt, gp
int	strsearch()
errchk	id_dbread, id_map, id_dbsave

begin
	verbose = clgetb ("verbose")
	debug = clgetb ("debug")

        dindef = INDEFD
	call amovkd (0.d0, wts, MAXPTS)

#  Read X and shift vectors from XCSAO output
	call clgstr ("shiftfile", shiftfile, SZ_FNAME)
	fd = open (shiftfile, READ_ONLY, TEXT_FILE)
	if (debug) {
	    call printf ("SKYCAL: Reading from shift file %s\n")
		call pargstr (shiftfile)
	    call flush (STDOUT)
	    verbose = TRUE
	    }
	npts = 0
	if (fd >0) {
	    while (fscan(fd) != EOF && npts < MAXPTS) {
		npts = npts + 1;
		call gargwrd (s1)
		call gargwrd (s2)
		call gargwrd (s3)
		call gargd (y[npts])
		call gargr (d1)
		call gargr (d2)
		call gargr (d3)
		call gargr (d4)
		call gargd (x[npts])
		wts[npts] = 1.0
		if (verbose) {
		    call printf ("%3d   %7.2f  %7.3f  %7.3f\n")
			call pargi (npts)
			call pargd (x[npts])
			call pargd (y[npts])
			call pargd (wts[npts])
		    }
		}
	    call close (fd)
	    if (verbose)
		call flush (STDOUT)
	    }

# Set the ICFIT pointer structure.
	call clgstr ("sfitfunc", fitfunc, SZ_FNAME)
	fitorder = clgeti ("sfitorder")
	niterate = 5
	hirej = 3.0
	lowrej = 3.0
	grow = 0
        call ic_open (ic)
        call ic_pstr (ic, "sample", "*")
        call ic_pstr (ic, "function", fitfunc)
        call ic_puti (ic, "naverage", 1)
        call ic_puti (ic, "order", fitorder)
        call ic_puti (ic, "niterate", niterate)
        call ic_putr (ic, "low", lowrej)
        call ic_putr (ic, "high", hirej)
        call ic_putr (ic, "grow", grow)
        call ic_pstr (ic, "ylabel", "")

#  Fit a function to X and Y (pixel and shift)
	if (verbose) {
	    call printf ("SKYCAL: %d points to fit\n")
		call pargi (npts)
	    call flush (STDOUT)
	    }

#  If the interactive flag is set then use icg_fit to set the fitting
#  parameters.  Only done if task is run interactively as well.
	interact = clgetb ("sinteract")
	if (interact) {
	    iferr {
		gp = gopen ("stdgraph", NEW_FILE, STDGRAPH)
		gt = gt_init ()
		call gt_sets (gt, GTTYPE, "line")
		} then
		    call error (0, "Error opening `stdgraph'.")
	    call icg_fitd (ic,gp,"cursor",gt,cv,x,y,wts,npts)
	    call gt_free (gt)
	    call gclose (gp)
	    }

#  Do the fit non-interactively using ic_fit
	else {
	    call ic_fitd (ic,cv,x,y,wts,npts,YES,YES,YES,YES)
	    }

#  Read reference image name
	call clgstr ("refin", refin, SZ_FNAME)
	if (debug) {
	    call printf ("SKYCAL: Input reference spectrum %s.\n")
		call pargstr (refin)
	    call flush (STDOUT)
	    }

# Allocate the basic data structures.
	call id_init (id)
	if (debug) {
	    call printf ("SKYCAL: ID structure initialized.\n")
	    call flush (STDOUT)
	    }
	call ic_open (ID_IC(id))
	if (debug) {
	    call printf ("SKYCAL: ID structure opened.\n")
	    call flush (STDOUT)
	    }

# Open the image and return if there is an error.
	call strcpy (refin, ID_IMAGE(id), ID_LENSTRING)
	iferr (call id_map (id)) {
	    call error (0, "Cannot read reference image")
	    return
	    }
	iferr (call id_dbsave (id, ID_IMAGE(id))) {
	    call printf ("SKYCAL: Cannot read reference image line IDs\n")
#	    call error (0, "Cannot read reference image line IDs")
#	    return
	    }
	if (debug) {
	    call printf ("SKYCAL: Image %s opened.\n")
		call pargstr (refin)
	    call flush (STDOUT)
	    }

# Get and save the reference database entry.
	call strcpy ("database", ID_DATABASE(id), ID_LENSTRING)
	call id_dbread (id, ID_IMAGE(id), ID_AP(id,1), NO, YES)
	call printf ("SKYCAL: Image database read.\n")
	call flush (STDOUT)
	call id_saveap (id)
	if (debug) {
	    call printf ("SKYCAL: Image database saved.\n")
	    call flush (STDOUT)
	    }

	nfeatures = ID_NFEATURES(id)
	if (nfeatures == 0)
	    call error (0, "No features in reference")
	else if (debug) {
	    call printf ("SKYCAL: %d features in reference.\n")
		call pargi (nfeatures)
	    call flush (STDOUT)
	    }

# For each reference feature a new pixel value is computed using
# the fit function
	call amovkd (0.d0, wts, MAXPTS)
	do i = 1, nfeatures {
	    dline0 = PIX(id,i)
	    if (debug) {
		call printf ("%2d: %.3f\n")
		    call pargi (i)
		    call pargd (dline0)
		call flush (STDOUT)
		}
		
	    PIX(id,i) = dline0 + dcveval (cv, dline0)
	    x[i] = PIX(id,i)
	    y[i] = USER(id,i)
	    WTS(id,i) = 1.0
	    wts[i] = 1.0
	    if (verbose) {
		call printf ("%2d: %.3f => %.3f %.3f %.3f %.3f\n")
		    call pargi (i)
		    call pargd (dline0)
		    call pargd (PIX(id,i))
		    call pargd (FIT(id,i))
		    call pargd (USER(id,i))
		    call pargd (WTS(id,i))
		}
	    }

# Set the ICFIT pointer structure.
	call clgstr ("dfitfunc", fitfunc, SZ_FNAME)
	fitorder = clgeti ("dfitorder")
	niterate = 5
	hirej = 3.0
	lowrej = 3.0
	grow = 0
        call ic_open (ID_IC(id))
        call ic_pstr (ID_IC(id), "sample", "*")
        call ic_pstr (ID_IC(id), "function", fitfunc)
        call ic_puti (ID_IC(id), "naverage", 1)
        call ic_puti (ID_IC(id), "order", fitorder)
        call ic_puti (ID_IC(id), "niterate", niterate)
        call ic_putr (ID_IC(id), "low", lowrej)
        call ic_putr (ID_IC(id), "high", hirej)
        call ic_putr (ID_IC(id), "grow", grow)

#  Now fit a function to pixels (X) and wavelength (Y)
	if (verbose) {
	    call printf ("SKYCAL: %d points to fit\n")
		call pargi (nfeatures)
	    do i = 1, nfeatures {
		call printf ("%3d   %7.2f  %7.3f  %5.3f\n")
		    call pargi (i)
		    call pargd (x[i])
		    call pargd (y[i])
		    call pargd (wts[i])
		}
	    call flush (STDOUT)
	    }

#  If the interactive flag is set then use icg_fit to set the fitting
#  parameters.  Only done if task is run interactively as well.
	interact = clgetb ("dinteract")
	if (interact) {
	    iferr {
		gp = gopen ("stdgraph", NEW_FILE, STDGRAPH)
		gt = gt_init ()
		call gt_sets (gt, GTTYPE, "line")
		} then
		    call error (0, "Error opening `stdgraph'.")
	    call icg_fitd (ic,gp,"cursor",gt,ID_CV(id),x,y,wts,nfeatures)
	    call gt_free (gt)
	    call gclose (gp)
	    }

#  Do the fit non-interactively using ic_fit
	else {
	    call ic_fitd (ID_IC(ID),ID_CV(id),x,y,wts,npts,YES,YES,YES,YES)
	    }

#  Recompute the fit wavelengths for the stretched and shifted pixels
	call id_fitdata (id)

	if (verbose) {
	    call printf ("\n i    Pix     Fit      User      Wt\n")
	    do i = 1, nfeatures {
		call printf ("%2d: %8.3f %8.3f %8.3f %5.3f\n")
		    call pargi (i)
		    call pargd (PIX(id,i))
		    call pargd (FIT(id,i))
		    call pargd (USER(id,i))
		    call pargd (WTS(id,i))
		}
	    call flush (STDOUT)
	    }

# Write revised pixel numbers to the output database
	call clgstr ("refout", refout, SZ_FNAME)
	call strcpy ("database/id", dbfile, SZ_FNAME)
	call strcat (refout, dbfile, SZ_FNAME)
	if (debug) {
	    call printf ("SKYCAL: Output reference spectrum %s.\n")
		call pargstr (refout)
	    call flush (STDOUT)
	    }
	call strcpy (refout, ID_IMAGE(id), ID_LENSTRING)
	ap[1] = 1
	ap[2] = 0
	ir = strsearch (refout, ".fits")
	if (ir > 0)
	    refout[ir-5] = EOS
	else {
	    ir = strsearch (refout, ".imh")
	    if (ir > 0)
		refout[ir-4] = EOS
	    }
	ID_NEWFEATURES(id) = YES
	ID_NEWCV(id) = YES
	ID_NEWDBENTRY(id) = YES
	call id_dbwrite (id, refout, ap, verbose)
	call show_sky (id, "STDOUT")

end


# SHOW_SKY -- Show parameter information. (from ID_SHOW)

procedure show_sky (id, file)

pointer	id			# ID pointer
char	file[ARB]		# File

char	line[SZ_LINE]
int	fd

int	open(), ic_geti()
errchk	open()

begin
	fd = open (file, APPEND, TEXT_FILE)

	call sysid (line, SZ_LINE)
	call fprintf (fd, "%s\n")
	call pargstr (line)
	call flush (fd)

	call fprintf (fd, "image %s\n")
	    call pargstr (ID_IMAGE(id))
	call fprintf (fd, "nsum %d\n")
	    call pargi (ID_NSUM(id,1))
	switch (ID_FTYPE(id)) {
	case EMISSION:
	    call fprintf (fd, "ftype emission\n")
	case ABSORPTION:
	    call fprintf (fd, "ftype absorption\n")
	}
	call flush (fd)
	switch (ID_LABELS(id)) {
	case 2:
	    call fprintf (fd, "labels index\n")
	case 3:
	    call fprintf (fd, "labels pixel\n")
	case 4:
	    call fprintf (fd, "labels coords\n")
	case 5:
	    call fprintf (fd, "labels user\n")
	case 6:
	    call fprintf (fd, "labels both\n")
	default:
	    call fprintf (fd, "labels none\n")
	}
	call flush (fd)
	call fprintf (fd, "maxfeatures %d\n")
	    call pargi (ID_MAXFEATURES(id))
	call flush (fd)
	call fprintf (fd, "match %g\n")
	    call pargr (ID_MATCH(id))
	call flush (fd)
	call fprintf (fd, "zwidth %g\n")
	    call pargr (ID_ZWIDTH(id))
	call flush (fd)
	call fprintf (fd, "fwidth %g\n")
	    call pargr (ID_FWIDTH(id))
	call flush (fd)
	call fprintf (fd, "database %s\n")
	    call pargstr (ID_DATABASE(id))
	call flush (fd)
	call fprintf (fd, "coordlist %s\n")
	    call pargstr (ID_COORDLIST(id))
	call flush (fd)
	call fprintf (fd, "cradius %g\n")
	    call pargr (ID_CRADIUS(id))
	call flush (fd)
	call fprintf (fd, "threshold %g\n")
	    call pargr (ID_THRESHOLD(id))
	call flush (fd)
	call fprintf (fd, "minsep %g\n")
	    call pargr (ID_MINSEP(id))
	call flush (fd)
	if (ID_CV(id) != NULL) {
	    call fprintf (fd, "function = %s\n")
	        call ic_gstr (ID_IC(id), "function", line, SZ_LINE)
		call pargstr (line)
	    call flush (fd)
	    call fprintf (fd, "order = %d\n")
		call pargi (ic_geti (ID_IC(id), "order"))
	    call flush (fd)
#	    call fprintf (fd, "Fit at first pixel = %0.8g\n")
#		call pargd (FITDATA(id,1))
#	    call flush (fd)
#	    call fprintf (fd, "Average fit interval = %0.8g\n")
#		call pargd ((FITDATA(id,ID_NPTS(id))-FITDATA(id,1))/
#		    (ID_NPTS(id)-1))
#	    call flush (fd)
	}

	call close (fd)
end

# May 21 2003	New program based on reref
